; optimized SHA1 (FIPS PUB 180-1) implementation 
;
; Copyright (c) 2003 Jan Wassenberg
;
; This program is free software; you can redistribute it and/or
; modify it under the terms of the GNU General Public License as
; published by the Free Software Foundation; either version 2 of the
; License, or (at your option) any later version.
;
; This program is distributed in the hope that it will be useful, but
; WITHOUT ANY WARRANTY; without even the implied warranty of
; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
; General Public License for more details.
;
; Contact info:
;   Jan.Wassenberg@stud.uni-karlsruhe.de
;   http://www.stud.uni-karlsruhe.de/~urkt/


format COFF


; NOTES
; - CPU >= P3/Athlon required (cmovcc, prefetch)
; - not reentrant!


public _sha1_update
public _sha1_finish
public _sha1

extrn __imp__CreateFileA@28
extrn __imp__GetFileSize@8
extrn __imp__ReadFileEx@20
extrn __imp__SleepEx@8
extrn __imp__CloseHandle@4
extrn __imp___aligned_malloc
extrn __imp___aligned_free


CSIZE		equ	262144				; read request size
							; (unbuffered i/o => = n*sector_size)
							; (reused as buf alignment => = 2^m)


;############################################################################################################################
; data
;############################################################################################################################


section '.data' data

org_H:
		dd	0x67452301
		dd	0xEFCDAB89
		dd	0x98BADCFE
		dd	0x10325476
		dd	0xC3D2E1F0


section '.bss'

W		rd	80
H		rd	5

bytes_in_W	rd	1				; allow partial block update
len		rd	2				; total bytes hashed (appended to msg) (64 bit)

_esp		rd	1				; saved esp from hash_block (uses all regs)

hFile		rd	1
bytes_left	rd	2
ovl		rd	5				; OVERLAPPED
buf		rd	4				; array of pointers to read buffers

next_read_buf	rd	1
next_hash_buf	rd	1

bss_size	equ	72				; size of 0-init vars (after H)


;############################################################################################################################
; macros
;############################################################################################################################

; f1(B, C, D) = (B & (C ^ D)) ^ D
macro r1 A, B, C, D, E, ofs
{
	mov		edi, C
	mov		esp, A
	add		E, [W+ebp+80+0+ofs]

	xor		edi, D
	rol		esp, 5

	and		edi, B
	lea		E, [esp+E+0x5A827999]

	xor		edi, D
	ror		B, 2

	add		E, edi
}

; f2(B, C, D) = (B ^ C ^ D)
macro r2 A, B, C, D, E, ofs
{
	mov		edi, B
	mov		esp, A
	add		E, [W+ebp+80+80+ofs]

	xor		edi, C
	rol		esp, 5

	xor		edi, D
	lea		E, [esp+E+0x6ED9EBA1]

	ror		B, 2

	add		E, edi
}

; f3(B, C, D) = ((B | C) & D) | (B & C)
macro r3 A, B, C, D, E, ofs
{
	mov		edi, B
	mov		esp, A
	add		E, [W+ebp+80+160+ofs]

	or		edi, C
	rol		esp, 5

	and		edi, D
	lea		E, [esp+E+0x8F1BBCDC]
	mov		esp, B

	ror		B, 2
	and		esp, C

	or		edi, esp

	add		E, edi
}

; f4(B, C, D) = (B ^ C ^ D)
macro r4 A, B, C, D, E, ofs
{
	mov		edi, B
	mov		esp, A
	add		E, [W+ebp+80+240+ofs]

	xor		edi, C
	rol		esp, 5

	xor		edi, D
	lea		E, [esp+E+0xCA62C1D6]

	ror		B, 2

	add		E, edi
}

; 20 iterations of subround i (A..E renamed cyclically)
macro r i
{
local l1

	mov		ebp, -20*4
l1:
r#i eax, ebx, ecx, edx, esi, 0
r#i esi, eax, ebx, ecx, edx, 4
r#i edx, esi, eax, ebx, ecx, 8
r#i ecx, edx, esi, eax, ebx, 12
r#i ebx, ecx, edx, esi, eax, 16
	add		ebp, 20
	jnz		l1
}


;############################################################################################################################
; code
;############################################################################################################################

section '.text' code


;ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ


;____________________________________________________________________________________________________________________________
;
; update digest with hash of a 64 byte block
;
; > esi -> data (little endian)
; x eax, ecx, edx, edi, ebp
;
; (called from _sha1_update, _sha1_finish)
;____________________________________________________________________________________________________________________________
;

hash_block:
	push		ebx
	push		esi
	mov		[_esp], esp

	mov		edi, W

; a. Divide Mi into 16 words W0, W1, ... , W15
;
;    (bswap to big endian)
	mov		ecx, -8
.l1:	mov		eax, [esi+ecx*8+64]
	mov		ebx, [esi+ecx*8+64+4]
	bswap		eax
	bswap		ebx
	mov		[edi+ecx*8+64], eax
	mov		[edi+ecx*8+64+4], ebx
	inc		ecx
	jnz		.l1

	add		edi, 64+64*4

; b. for t = 16 to 79 do
;      Wt = S1(Wt-3 XOR Wt-8 XOR Wt-14 XOR Wt-16).
;
;    (faster than circular buffer method)
	mov		ecx, -64
.l2:	mov		eax, [edi+ecx*4-3*4]
	mov		ebx, [edi+ecx*4-8*4]
	xor		eax, [edi+ecx*4-14*4]
	xor		ebx, [edi+ecx*4-16*4]
	xor		eax, ebx
	inc		ecx
	rol		eax, 1
	mov		[edi+ecx*4-4], eax
	jnz		.l2

; c. A = H0, B = H1, C = H2, D = H3, E = H4
	mov		eax, [H+0]
	mov		ebx, [H+4]
	mov		ecx, [H+8]
	mov		edx, [H+12]
	mov		esi, [H+16]

; d. for t = 0 to 79 do
;      TEMP = S5(A) + ft(B,C,D) + E + Wt + Kt;
;      E = D; D = C; C = S30(B); B = A; A = TEMP;
r 1
r 2
r 3
r 4

; e. H0 += A, H1 += B, H2 += C, H3 += D, H4 += E
	add		[H+0], eax
	add		[H+4], ebx
	add		[H+8], ecx
	add		[H+12], edx
	add		[H+16], esi

	mov		esp, [_esp]
	pop		esi
	pop		ebx
	ret


;____________________________________________________________________________________________________________________________
;
; extern void sha1_update(unsigned char* data, unsigned long len);
;____________________________________________________________________________________________________________________________
;

_sha1_update:
	pusha

	mov		esi, [esp+32+4]			; data
	mov		ebx, [esp+32+8]			; len

	xor		eax, eax
	add		[len], ebx
	adc		[len+4], eax

; fill buf, if possible
	mov		ebp, [bytes_in_W]
	test		ebx, ebx
	jz		.end				; protect rep from 0 count
	lea		edi, [W+ebp]			; end of W
	mov		edx, 64
	sub		edx, ebp			; avl
	mov		ecx, ebx
	add		ebp, ebx			; end += len (guess)
	sub		ebx, edx			; len -= avl (guess)
	cmovge		ecx, edx			; trans_size = min(len, avl)
rep	movsb
	jl		.end				; was partial block: exit
	push		esi
	mov		esi, W
	call		hash_block
	pop		esi

; hash complete blocks
	push		ebx				; save for partial block below
	shr		ebx, 6				; # blocks
	jz		.no_block
.block_loop:
	prefetcht0	[esi+128]
	call		hash_block
	add		esi, 64
	dec		ebx
	jnz		.block_loop
.no_block:

; copy remainder into W
	pop		ebp
	and		ebp, 0x3f
	jz		.end				; protect rep from 0 count
	mov		ecx, ebp
	mov		edi, W
rep	movsb
.end:

	mov		[bytes_in_W], ebp

	popa
	ret


;____________________________________________________________________________________________________________________________
;
; extern void sha1_finish(unsigned char* digest /* [20] */);
;____________________________________________________________________________________________________________________________
;

_sha1_finish:
	pusha

; pad message to 448 (mod 512) bits
	mov		edx, [bytes_in_W]
	mov		ecx, 64
	sub		ecx, edx
	lea		edi, [W+edx]
	mov		al, 0x80
	stosb
	xor		eax, eax
rep	stosb						; append 0s

; hash current block, start another (if necessary)
	cmp		edx, 55				; need 2nd block? (length doesn't fit)
	jle		.no_2nd_block			; no - skip

	mov		esi, W
	call		hash_block

	mov		edi, W
	xor		eax, eax
	mov		ecx, 64/4
rep	stosd						; zero block

.no_2nd_block:

; append message length (bits)
	mov		eax, [len]
	mov		edx, [len+4]
	shld		edx, eax, 3
	shl		eax, 3
	bswap		eax
	bswap		edx
	mov		[W+14*4], edx
	mov		[W+15*4], eax

	mov		esi, W
	call		hash_block

; copy digest
	mov		esi, H
	mov		edi, [esp+32+4]			; digest
times 5	movsd

	popa
	ret


;ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ


;____________________________________________________________________________________________________________________________
;
; start read of next chunk (async)
;
; x eax, ecx, edx, esi, edi
;
; (called from _read_complete_cb@12, _sha1)
;____________________________________________________________________________________________________________________________
;

start_next_read:

; read_buf = buf[next_read_buf];
	mov		eax, [next_read_buf]
	mov		edx, [buf+eax*4]

; next_read_buf = (next_read_buf+1) & 3;
	inc		eax
	and		eax, 3
	mov		[next_read_buf], eax

	mov		esi, ovl			; reused
	mov		edi, CSIZE			;

; ReadFileEx(hFile, buf[cur_read], CSIZE, &ovl, read_complete_cb);
	push		_read_complete_cb@12
	push		esi
	push		edi
	push		edx
	push		dword [hFile]
	call		dword [__imp__ReadFileEx@20]

; ovl.Offset += CSIZE;
	xor		eax, eax
	add		[esi+8], edi
	adc		[esi+12], eax

	ret


;____________________________________________________________________________________________________________________________
;
; finished reading a chunk: start reading next chunk, hash current chunk
;
; (I/O completion routine)
;____________________________________________________________________________________________________________________________
;

_read_complete_cb@12:
	pusha

	call		start_next_read

; pos = buf[next_hash_buf]
	mov		eax, [next_hash_buf]
	mov		esi, [buf+eax*4]

; next_hash_buf = (next_hash_buf+1) & 3
	inc		eax
	and		eax, 3
	mov		[next_hash_buf], eax

	mov		ecx, [esp+32+8]			; bytes_read
	xor		eax, eax
	sub		[bytes_left], ecx
	sbb		[bytes_left+4], eax

	push		ecx
	push		esi
	call		_sha1_update
	pop		edx				; cdecl call
	pop		edx

	popa
	ret		12


;____________________________________________________________________________________________________________________________
;
; int __stdcall sha1(char* filename, unsigned char* digest /* [20] */)
;
; return:
;   -1: file not found
;    0: not enough mem
;    1: success; 20 byte message digest stored in digest (must be valid pointer)
;____________________________________________________________________________________________________________________________
;

_sha1:
	push		esi				; return eax => don't pusha
	push		edi

; init H
	mov		esi, org_H
	mov		edi, H
	push		5
	pop		ecx
rep	movsd

; zero init variables
	xor		eax, eax
	mov		cl, bss_size/4
rep	stosd

; hFile = CreateFile(filename, GENERIC_READ, 0, 0, OPEN_EXISTING, FILE_FLAG_OVERLAPPED|FILE_FLAG_NO_BUFFERING|FILE_FLAG_SEQUENTIAL_SCAN, 0);
	push		eax
	push		0x68000000			; FILE_FLAG_OVERLAPPED|FILE_FLAG_NO_BUFFERING|FILE_FLAG_SEQUENTIAL_SCAN
	push		3				; OPEN_EXISTING
	push		eax
	push		eax
	push		0x80000000			; GENERIC_READ
	push		dword [esp+8+24+4]		; filename
	call		dword [__imp__CreateFileA@28]

	cmp		eax, -1
	je		.err

	mov		[hFile], eax

; int len = GetFileSize(hFile, &bytes_left.hi);
	push		bytes_left+4
	push		eax
	call		dword [__imp__GetFileSize@8]
	mov		[bytes_left], eax

; alloc 4 bufs (unbuffered async read => mem must be sector aligned)
	mov		esi, CSIZE			; reused below
	push		esi
	push		CSIZE*4+256			; block loop prefetches past EOF
	call		dword [__imp___aligned_malloc]
	pop		edx				; __cdecl call
	pop		edx
	test		eax, eax
	jz		.err
	mov		edi, buf
repeat 4
	stosd
	add		eax, esi
end repeat

	call		start_next_read
	call		start_next_read

; while(chunks > 0)
;   SleepEx(INFINITE, TRUE);
.l3:
	push		1
	push		-1
	call		dword [__imp__SleepEx@8]
	cdq						; SleepEx returns 0xc0 => edx = 0
	cmp		[bytes_left], edx
	ja		.l3
	cmp		[bytes_left+4], edx
	ja		.l3

	push		dword [esp+8+8]			; digest
	call		_sha1_finish
	pop		edx				; cdecl call

; close file
	push		dword [hFile]
	call		dword [__imp__CloseHandle@4]

; free mem
	push		dword [buf]
	call		dword [__imp___aligned_free]
	pop		edx				; __cdecl call

	xor		eax, eax
	inc		eax				; return success
.err:
	pop		edi
	pop		esi

	ret