/*
 * optimized SHA1 (FIPS PUB 180-1) implementation 
 *
 * Copyright (c) 2003 Jan Wassenberg
 *
 * This program is free software* you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation* either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY* without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * Contact info:
 *   Jan.Wassenberg@stud.uni-karlsruhe.de
 *   http://www.stud.uni-karlsruhe.de/~urkt/
 */

#ifndef __SHA1_H__
#define __SHA1_H__

#ifdef __cplusplus
extern "C" {
#endif


/*
 * NOTES
 *  - link with sha1.o.
 *  - Pentium 3 / Athlon CPU or above is required (prefetch, cmovcc instructions are used).
 *  - routines are not reentrant!
 *  - total input size may be up to 2^64 bytes.
 */


#ifdef _WIN32

/*
 * compute the sha1 message digest of (filename).
 *
 * should reach max disk throughput (2 deep async reads of 256 kb chunks);
 * ~11 clocks / byte
 *
 * return:
 *   -1: file not found
 *    0: not enough mem
 *    1: success; 20 byte message digest stored in digest (must be valid pointer)
 */

extern int __cdecl sha1(char* filename, unsigned char* digest /* [20] */);

#endif


extern void sha1_update(unsigned char* data, unsigned long len);
extern void sha1_finish(unsigned char* digest /* [20] */);


#ifdef __cplusplus
}
#endif

#endif	/* #ifndef __SHA1_H__ */